#include "tkInt.h"
#include "SdlTkInt.h"

/*
 * Undocumented Xlib internal function
 */
int _XInitImageFuncPtrs(XImage *image)
{
    return 0;
}

XClassHint *
XAllocClassHint(void)
{
    XClassHint *hint;

    hint = (XClassHint *) ckalloc(sizeof(XClassHint));
    return hint;
}

int
XAllocColor(display, colormap, color)
    Display* display;
    Colormap colormap;
    XColor* color;
{
    /* NOTE: If this changes, update TkpGetPixel */
    Uint8 r = (color->red / 65535.0) * 255.0;
    Uint8 g = (color->green / 65535.0) * 255.0;
    Uint8 b = (color->blue / 65535.0) * 255.0;

    display->request++;
    color->pixel = SDL_MapRGB(sdl_screen->format, r, g, b);
    return 1;
}

Status
XAllocNamedColor(display, colormap, color_name, screen_def_return,
	exact_def_return)
    Display* display;
    Colormap colormap;
    CONST char* color_name;
    XColor* screen_def_return;
    XColor* exact_def_return;
{
    /* xcolors.c */
    if (XParseColor(display, colormap, color_name, exact_def_return) == 1) {
	*screen_def_return = *exact_def_return;
	return XAllocColor(display, colormap, screen_def_return);
    }
    return 0;
}

XSizeHints *
XAllocSizeHints(void)
{
    return NULL;
}

void
XBell(display, percent)
    Display* display;
    int percent;
{
}

void
XChangeGC(d, gc, mask, values)
    Display * d;
    GC gc;
    unsigned long mask;
    XGCValues *values;
{
    d->request++;

    if (mask & GCFunction) { gc->function = values->function; }
    if (mask & GCPlaneMask) { gc->plane_mask = values->plane_mask; }
    if (mask & GCForeground) { gc->foreground = values->foreground; }
    if (mask & GCBackground) { gc->background = values->background; }
    if (mask & GCLineWidth) { gc->line_width = values->line_width; }	
    if (mask & GCLineStyle) { gc->line_style = values->line_style; }
    if (mask & GCCapStyle) { gc->cap_style = values->cap_style; }
    if (mask & GCJoinStyle) { gc->join_style = values->join_style; }
    if (mask & GCFillStyle) { gc->fill_style = values->fill_style; }
    if (mask & GCFillRule) { gc->fill_rule = values->fill_rule; }
    if (mask & GCArcMode) { gc->arc_mode = values->arc_mode; }
    if (mask & GCTile) { gc->tile = values->tile; }
    if (mask & GCStipple) { gc->stipple = values->stipple; }
    if (mask & GCTileStipXOrigin) { gc->ts_x_origin = values->ts_x_origin; }
    if (mask & GCTileStipYOrigin) { gc->ts_y_origin = values->ts_y_origin; }
    if (mask & GCFont) { gc->font = values->font; }
    if (mask & GCSubwindowMode) { gc->subwindow_mode = values->subwindow_mode; }
    if (mask & GCGraphicsExposures) { gc->graphics_exposures = values->graphics_exposures; }
    if (mask & GCClipXOrigin) { gc->clip_x_origin = values->clip_x_origin; }
    if (mask & GCClipYOrigin) { gc->clip_y_origin = values->clip_y_origin; }
    if (mask & GCClipMask) { XSetClipMask(d, gc, values->clip_mask); }
    if (mask & GCDashOffset) { gc->dash_offset = values->dash_offset; }
    if (mask & GCDashList) { gc->dashes = values->dashes; (&(gc->dashes))[1] = 0;}

dbwin("CHANGE GC: %p\n", (void *) gc);
}

void
XChangeProperty(display, w, property, type, format, mode, data, nelements)
    Display* display;
    Window w;
    Atom property;
    Atom type;
    int format;
    int mode;
    _Xconst unsigned char* data;
    int nelements;
{
    _Window *_w = (_Window *) w;

    /* UpdateTitle() */
    if ((property > XA_LAST_PREDEFINED) &&
	    !strcmp((char *) property, "_NET_WM_NAME")) {

	if (_w->title != NULL)
	    ckfree((char *) _w->title);
	_w->title = ckalloc(nelements + 1); /* UTF-8 */
	strcpy((char *) _w->title, (char *) data);

	/* Redraw frame titlebar */
	if (_w->parent->dec != NULL) {
	    decSetDraw(_w->parent, 1);
	    SdlTkScreenChanged(0);
	}
    }
    /* FIXME: _NET_WM_ICON_NAME as well */
}

void
XChangeWindowAttributes(display, w, valueMask, attributes)
    Display* display;
    Window w;
    unsigned long valueMask;
    XSetWindowAttributes* attributes;
{
    _Window *_w = (_Window *) w;

    if (valueMask & CWCursor) {
	XDefineCursor(display, w, attributes->cursor);
    }
    if (valueMask & CWEventMask) {
	_w->atts.your_event_mask = attributes->event_mask;
    }
    if (valueMask & CWOverrideRedirect) {
	if (dbg & DBG_WINDOWS) dbwin("XChangeWindowAttributes %s override_redirect=%d\n", SdlTkDescribeWindow(w), attributes->override_redirect);

	/* Tk won't call us unless it changed */
	_w->atts.override_redirect = attributes->override_redirect;

	/* Is override_redirect, wasn't before */
	if (attributes->override_redirect) {

	    /* Decorative frame may not have been allocated yet
	     * if the window was never mapped */
	    if (_w->parent->dec != NULL) {
		Window wdec = (Window) _w->parent;

		/* Reparent to root */
		XReparentWindow(display, w, x_screen->root,
		    _w->parent->atts.x,
		    _w->parent->atts.y);

		/* Destroy decorative frame */
		XDestroyWindow(display, wdec);
	    }

	    /* This must be the Tk wrapper */
	    if (PARENT_IS_ROOT(_w)) {
		SdlTkResizeWindowSurface(_w);
	    }

	/* Was override_redirect, isn't now */
	} else {
	    if (_w->sdl != NULL) {
		if (dbg & DBG_SURFACE) dbwin("XChangeWindowAttributes: freeing surface of %s\n", SdlTkDescribeWindow(w));
		SDL_FreeSurface(_w->sdl);
		_w->sdl = NULL;
	    }
	    XUnmapWindow(display, w);
	    XMapWindow(display, w);
	}
    }
}

/*
void
XClipBox(r, rect_return)
    Region r;
    XRectangle* rect_return;
{
}
*/
int
XCloseDisplay(display)
    Display* display;
{
    return 0;
}

void
XConfigureWindow(display, w, value_mask, values)
    Display* display;
    Window w;
    unsigned int value_mask;
    XWindowChanges* values;
{
    _Window *_w = (_Window *) w;

    display->request++;

    /* I don't think this border_width is ever used, so hard to test it */
    /* A widget's -borderwidth option is completely different */
    if (value_mask & CWBorderWidth) {
	_w->atts.border_width = values->border_width;
	_w->parentWidth = _w->atts.width + 2 * values->border_width;
	_w->parentHeight = _w->atts.height + 2 * values->border_width;
	SdlTkScreenChanged(1);
    }

    /* Need this for Tk_RestackWindow and Tk_MakeWindowExist */
    if (value_mask & CWStackMode) {
	_Window *sibling = NULL;

	if (value_mask & CWSibling)
	    sibling = (_Window *) values->sibling;

	SdlTkRestackWindow(_w, sibling, values->stack_mode);
	/*SdlTkRestackTransients(_w); -- this isn't a toplevel */

	SdlTkScreenChanged(1);
    }
}

int
XConvertSelection(display, selection, target, property, requestor, time)
    Display* display;
    Atom selection;
    Atom target;
    Atom property;
    Window requestor;
    Time time;
{
    return 0;
}

void
XCopyArea(display, src, dest, gc, src_x, src_y, width, height, dest_x, dest_y)
    Display* display;
    Drawable src;
    Drawable dest;
    GC gc;
    int src_x, src_y;
    unsigned int width, height;
    int dest_x, dest_y;
{
    if (dbg & DBG_DRAW) dbwin("XCopyArea src=%p dest=%p x=%d y=%d width=%d height=%d dest_x=%d dest_y=%d\n", (void *) src, (void *) dest, src_x, src_y, width, height, dest_x, dest_y);

    display->request++;

    gfxCopyArea(src, dest, gc, src_x, src_y, width, height, dest_x, dest_y);

    if (IS_WINDOW(dest)) {
	TkpClipMask *clipPtr = (TkpClipMask *) gc->clip_mask;
	SdlTkScreenChanged(0);
	if (clipPtr && clipPtr->type == TKP_CLIP_REGION) {
	    Region clipRgn = (Region) clipPtr->value.region;
	    SdlTkDirtyRegion(dest, clipRgn);
	} else {
	    SdlTkDirtyArea(dest, dest_x, dest_y, width, height);
	}
    }
}

void
XCopyPlane(display, src, dest, gc, src_x, src_y, width, height, dest_x,
	dest_y, plane)
    Display* display;
    Drawable src;
    Drawable dest;
    GC gc;
    int src_x, src_y;
    unsigned int width, height;
    int dest_x, dest_y;
    unsigned long plane;
{
    if (dbg & DBG_DRAW) dbwin("XCopyPlane src=%p dest=%p x=%d y=%d width=%d height=%d dest_x=%d dest_y=%d\n", (void *) src, (void *) dest, src_x, src_y, width, height, dest_x, dest_y);

    display->request++;

    gfxDrawBitmap(src, dest, gc, src_x, src_y, width, height, dest_x,
	dest_y);

    if (IS_WINDOW(dest))
	SdlTkScreenChanged(0);
}

Pixmap
XCreateBitmapFromData(display, d, data, width, height)
    Display* display;
    Drawable d;
    _Xconst char* data;
    unsigned int width;
    unsigned int height;
{
    XImage ximage;
/*    GC gc;*/
    Pixmap pix;
#if 1
    _Pixmap *_p;
    SDL_Surface *sdl;
    SDL_Color colors[2];

    /* Use 1 byte-per-pixel for efficient drawing/stippling */
    sdl = SDL_CreateRGBSurface(SDL_SWSURFACE, width, height,
	8, 0, 0, 0, 0);
    if (sdl == NULL)
	return None;

    /* New 8-bit surfaces have an empty palette. Set the palette to black
     * and white. */
    colors[0].r = colors[0].g = colors[0].b = 0;
    colors[1].r = colors[1].g = colors[1].b = 255;
    SDL_SetPalette(sdl, SDL_LOGPAL, colors, 0, 2);

    _p = (_Pixmap *) ckalloc(sizeof(_Pixmap));
    _p->type = DT_PIXMAP;
    _p->sdl = sdl;
    _p->format = SdlTkPixelFormat(sdl);

    pix = (Pixmap) _p;
#else

    /* FIXME: SDL bombs if I create a 1-bpp surface here (SDL_blit_0.c:BlitBto1 bug?) */
    pix = XCreatePixmap(display, d, width, height, 2);
#endif
/*    if (! (gc = XCreateGC(display, pix, (unsigned long) 0, (XGCValues *) 0)))
	return (Pixmap) None;*/
    ximage.height = height;
    ximage.width = width;
    ximage.depth = 1;
    ximage.bits_per_pixel = 1;
    ximage.xoffset = 0;
    ximage.format = XYPixmap;
    ximage.data = (char *)data;
    ximage.byte_order = LSBFirst;
    ximage.bitmap_unit = 8;
    ximage.bitmap_bit_order = LSBFirst;
    ximage.bitmap_pad = 8;
    ximage.bytes_per_line = (width+7)/8;

    XPutImage(display, pix, None /*gc*/, &ximage, 0, 0, 0, 0, width, height);
/*    XFreeGC(display, gc);*/
    return(pix);
}

Colormap
XCreateColormap(display, w, visual, alloc)
    Display* display;
    Window w;
    Visual* visual;
    int alloc;
{
    _Colormap *_colormap;

    display->request++;

    _colormap = (_Colormap *) ckalloc(sizeof(_Colormap));
    _colormap->whatever = 1234;

    return (Colormap) _colormap;
}

GC
XCreateGC(display, d, mask, values)
    Display* display;
    Drawable d;
    unsigned long mask;
    XGCValues* values;
{
    GC gp;

    display->request++;

/*
 * In order to have room for a dash list, MAX_DASH_LIST_SIZE extra chars are
 * defined, which is invisible from the outside. The list is assumed to end
 * with a 0-char, so this must be set explicitely during initialization.
 */

#define MAX_DASH_LIST_SIZE 10

    gp = (XGCValues *)ckalloc(sizeof(XGCValues) + MAX_DASH_LIST_SIZE);
    if (!gp) {
	return None;
    }

    gp->function = 	(mask & GCFunction) 	?values->function	:GXcopy;
    gp->plane_mask = 	(mask & GCPlaneMask) 	?values->plane_mask 	:~0;
    gp->foreground = 	(mask & GCForeground) 	?values->foreground 	:0;
    gp->background = 	(mask & GCBackground) 	?values->background 	:0xffffff;
    gp->line_width = 	(mask & GCLineWidth)	?values->line_width	:1;	
    gp->line_style = 	(mask & GCLineStyle)	?values->line_style	:LineSolid;
    gp->cap_style =  	(mask & GCCapStyle)	?values->cap_style	:0;
    gp->join_style = 	(mask & GCJoinStyle)	?values->join_style	:0;
    gp->fill_style =  	(mask & GCFillStyle)	?values->fill_style	:FillSolid;
    gp->fill_rule =  	(mask & GCFillRule)	?values->fill_rule	:WindingRule;
    gp->arc_mode = 	(mask & GCArcMode)	?values->arc_mode	:ArcPieSlice;
    gp->tile = 		(mask & GCTile)		?values->tile		:None;
    gp->stipple = 	(mask & GCStipple)	?values->stipple	:None;
    gp->ts_x_origin = 	(mask & GCTileStipXOrigin)	?values->ts_x_origin:0;
    gp->ts_y_origin = 	(mask & GCTileStipYOrigin)	?values->ts_y_origin:0;
    gp->font = 		(mask & GCFont)		?values->font		:None;
    gp->subwindow_mode = (mask & GCSubwindowMode)?values->subwindow_mode:ClipByChildren;
    gp->graphics_exposures = (mask & GCGraphicsExposures)?values->graphics_exposures:True;
    gp->clip_x_origin = (mask & GCClipXOrigin)	?values->clip_x_origin	:0;
    gp->clip_y_origin = (mask & GCClipYOrigin)	?values->clip_y_origin	:0;
    gp->dash_offset = 	(mask & GCDashOffset)	?values->dash_offset	:0;
    gp->dashes = 	(mask & GCDashList)	?values->dashes		:4;
    (&(gp->dashes))[1] = 	0;

    if (mask & GCClipMask) {
	gp->clip_mask = (Pixmap)ckalloc(sizeof(TkpClipMask));
	((TkpClipMask*)gp->clip_mask)->type = TKP_CLIP_PIXMAP;
	((TkpClipMask*)gp->clip_mask)->value.pixmap = values->clip_mask;
    } else {
	gp->clip_mask = None;
    }
/*dbwin("CREATE GC: %p\n", gp);*/
    return gp;
}

Cursor
XCreateGlyphCursor(display, source_font, mask_font, source_char, mask_char,
	foreground_color, background_color)
    Display* display;
    Font source_font;
    Font mask_font;
    unsigned int source_char;
    unsigned int mask_char;
    XColor* foreground_color;
    XColor* background_color;
{
    _Cursor *_c = (_Cursor *) ckalloc(sizeof(_Cursor));

    display->request++;

    return (Cursor) _c;
}

XIC
XCreateIC()
{
    return NULL;
}

XImage *
XCreateImage(display, visual, depth, format, offset, data, width, height,
	bitmap_pad, bytes_per_line)
    Display* display;
    Visual* visual;
    unsigned int depth;
    int format;
    int offset;
    char* data;
    unsigned int width;
    unsigned int height;
    int bitmap_pad;
    int bytes_per_line;
{
    XImage *ximage = (XImage *) ckalloc(sizeof(XImage));

    display->request++;

    /* Tk expects images to use the screen's pixel format */

    ximage->height = height;
    ximage->width = width;
    ximage->depth = depth;
    ximage->xoffset = offset;
    ximage->format = format;
    ximage->data = data;
    ximage->bitmap_pad = bitmap_pad;
    if (bytes_per_line == 0) {
	ximage->bytes_per_line = width * sdl_screen->format->BytesPerPixel;
    } else {
	ximage->bytes_per_line = bytes_per_line;
    }

    if (format == ZPixmap) {
	ximage->bits_per_pixel = sdl_screen->format->BitsPerPixel;
	ximage->bitmap_unit = sdl_screen->format->BitsPerPixel;
    } else {
	ximage->bits_per_pixel = 1;
	ximage->bitmap_unit = 8;
    }
    ximage->byte_order = LSBFirst;
    ximage->bitmap_bit_order = LSBFirst;
    ximage->red_mask = visual->red_mask;
    ximage->green_mask = visual->green_mask;
    ximage->blue_mask = visual->blue_mask;

    ximage->obdata = NULL;
    ximage->f.destroy_image = ImageDestroy;
    ximage->f.get_pixel = ImageGetPixel;
    ximage->f.put_pixel = ImagePutPixel;
    ximage->f.sub_image = NULL;
    ximage->f.add_pixel = NULL;

    return ximage;
}

Pixmap
XCreatePixmap(display, d, width, height, depth)
    Display* display;
    Drawable d;
    unsigned int width;
    unsigned int height;
    unsigned int depth;
{
    _Pixmap *_p;
    SDL_Surface *sdl;

    display->request++;

    if (dbg & DBG_PIXMAPS) dbwin("XCreatePixmap width=%d height=%d depth=%d", width, height, depth);

    /* Note: ignoring requested depth */
    /* Would like to support depth of 1 for creating bitmaps */
    sdl = SDL_CreateRGBSurface(SDL_SWSURFACE, width, height,
	(depth == 1) ? depth : sdl_screen->format->BitsPerPixel,
	sdl_screen->format->Rmask,
	sdl_screen->format->Gmask,
	sdl_screen->format->Bmask,
	sdl_screen->format->Amask);
    if (sdl == NULL)
	return None;

    _p = (_Pixmap *) ckalloc(sizeof(_Pixmap));
    _p->type = DT_PIXMAP;
    _p->sdl = sdl;
    _p->format = SdlTkPixelFormat(sdl);
    return (Pixmap) _p;
}

Cursor
XCreatePixmapCursor(display, source, mask, foreground_color,
	background_color, x, y)
    Display* display;
    Pixmap source;
    Pixmap mask;
    XColor* foreground_color;
    XColor* background_color;
    unsigned int x;
    unsigned int y;
{
    return (Cursor) NULL;
}
/*
Region XCreateRegion(void)
{
    return None;
}
*/
Window
XCreateWindow(display, parent, x, y, width, height, border_width,
    depth, class, visual, valuemask, attributes)
    Display* display;
    Window parent;
    int x;
    int y;
    unsigned int width;
    unsigned int height;
    unsigned int border_width;
    int depth;
    unsigned int class;
    Visual* visual;
    unsigned long valuemask;
    XSetWindowAttributes* attributes;
{
    _Window *_parent = (_Window *) parent;
    _Window *_w;

    display->request++;

    _w = (_Window *) ckalloc(sizeof(_Window));
    memset(_w, '\0', sizeof(_Window));

    _w->type = DT_WINDOW;
    _w->parent = (_Window *) parent;
    _w->atts.x = x;
    _w->atts.y = y;
    _w->atts.width = width;
    _w->atts.height = height;
    _w->atts.border_width = border_width;
    _w->atts.visual = visual;
    _w->atts.map_state = IsUnmapped;
    _w->atts.override_redirect = attributes ?
	attributes->override_redirect : False;
    _w->atts.your_event_mask = attributes ? attributes->event_mask : 0;

    /* A window's requested width/height are *inside* its borders.
     * It doesn't look like Tk uses this border_width anywhere, it is
     * different than the -borderwidth option. */
    _w->parentWidth = width + 2 * border_width;
    _w->parentHeight = height + 2 * border_width;

    _w->visRgn = RgnPoolGet();
    _w->visRgnInParent = RgnPoolGet();
    _w->dirtyRgn = RgnPoolGet();
    if (doublebuffer_windows)
	_w->visRgnInParent2 = RgnPoolGet();

    /* override_redirect windows get a surface */
    if (_w->atts.override_redirect)
	SdlTkResizeWindowSurface(_w);

    /* Make first child of parent */
    _w->next = _parent->child;
    _parent->child = _w;

    if (dbg & DBG_WINDOWS) dbwin("added child %s to parent %s\n",
	SdlTkDescribeWindow((Window)_w), SdlTkDescribeWindow(parent));

    return (Window) _w;
}

/* tkPointer.c */
#if 0
void
XDefineCursor(d, w, c)
    Display* d;
    Window w;
    Cursor c;
{
}
#endif

void
XDeleteProperty(display, w, property)
    Display* display;
    Window w;
    Atom property;
{
    _Window *_w = (_Window *) w;

    if (property == XA_WM_TRANSIENT_FOR) {
dbwin("%s master is now NULL\n", SdlTkDescribeWindow(w));
	_w->master = NULL;
    }
}

void
XDestroyIC(ic)
    XIC ic;
{
}
/*
void XDestroyRegion(Region r)
{
}
*/
void
XDestroyWindow(display, w)
    Display* display;
    Window w;
{
    _Window *_w = (_Window *) w;
    _Window *wdec = _w->parent->dec ? _w->parent : NULL;
    XEvent event;
    int doNotify = (_w->atts.your_event_mask & StructureNotifyMask) != 0;

    if (dbg & DBG_WINDOWS) dbwin("XDestroyWindow %s\n", SdlTkDescribeWindow(w));

    if (_w->atts.map_state != IsUnmapped)
	XUnmapWindow(display, w);

    display->request++;

    /* Destroy children recursively */
    while (_w->child != NULL)
	XDestroyWindow(display, (Window) _w->child);

    /* tkPointer.c cleanup (unless this is a decorative frame) */
    if (_w->tkwin != NULL)
	TkPointerDeadWindow(_w->tkwin);
    
    /* Remove from parent */
    SdlTkRemoveFromParent(_w);

/*if (x_focus_window == w) SdlTkLostFocusWindow();*/

    /* Free the toplevel drawing surface */
    if (_w->sdl != NULL) {
	SDL_FreeSurface(_w->sdl);
    }

    /* Free the decorative frame record */
    if (_w->dec != NULL) {
	decDestroy(_w);
    }

    if (_w->title != NULL)
	ckfree((char *) _w->title);

    if (_w->pathName != NULL)
	ckfree(_w->pathName);

    RgnPoolFree(_w->visRgn);
    RgnPoolFree(_w->visRgnInParent);
    RgnPoolFree(_w->dirtyRgn);
    if (doublebuffer_windows)
	RgnPoolFree(_w->visRgnInParent2);

    memset(_w, 0xFE, sizeof(_Window));
    ckfree((char *) _w);

    /* FIXME: Because the Window id is actually a pointer, another window
     * created later may occupy the same address as this deleted window.
     * So this DestroyNotify may be delivered to the new window.
     */
    /* I'm not sure this is needed. Tk generates its own DestroyNotify
     * events. */
#if 0
    if (doNotify) {
	event.type = DestroyNotify;
	event.xdestroywindow.serial = LastKnownRequestProcessed(display);
	event.xdestroywindow.send_event = False;
	event.xdestroywindow.display = display;
	event.xdestroywindow.event = w;
	event.xdestroywindow.window = w;
	Tk_QueueWindowEvent(&event, TCL_QUEUE_TAIL);
    }
#endif
#if 1
    /* Destroy decorative frame */
    /* Actually this shouldn't happen, since Tk reparents the wrapper
     * to the root before destroying it, and reparenting should destroy the
     * decorative frame ---> fix XReparentWindow to destroy decframe */
    if (wdec != NULL) {
	XDestroyWindow(display, (Window) wdec);
    }
#endif
    SdlTkScreenChanged(1);
}

void
XDrawArc(display, d, gc, x, y, width, height, start, extent)
    Display* display;
    Drawable d;
    GC gc;
    int x;
    int y;
    unsigned int width;
    unsigned int height;
    int start;
    int extent;
{
    display->request++;

    gfxDrawArc(d, gc, x, y, width, height, start, extent);

    if (IS_WINDOW(d)) {
	SdlTkScreenChanged(0);
	SdlTkDirtyAll(d);
    }
}

void
XDrawLine(display, d, gc, x1, y1, x2, y2)
    Display* display;
    Drawable d;
    GC gc;
    int x1, y1, x2, y2;		/* Coordinates of line segment. */
{
    XPoint points[2];

    points[0].x = x1;
    points[0].y = y1;
    points[1].x = x2;
    points[1].y = y2;
    XDrawLines(display, d, gc, points, 2, CoordModeOrigin);
}

void
XDrawLines(display, d, gc, points, npoints, mode)
    Display* display;
    Drawable d;
    GC gc;
    XPoint* points;
    int npoints;
    int mode;
{
    display->request++;

    gfxDrawLines(d, gc, points, npoints, mode);

    if (IS_WINDOW(d)) {
	SdlTkScreenChanged(0);
	SdlTkDirtyAll(d);
    }
}

void
XDrawPoint(display, d, gc, x, y)
    Display *display;
    Drawable d;
    GC gc;
    int x, y;
{
    gfxDrawPoint(d, gc, x, y);
}

void
XDrawRectangle(display, d, gc, x, y, width, height)
    Display* display;
    Drawable d;
    GC gc;
    int x;
    int y;
    unsigned int width;
    unsigned int height;
{
    if (dbg & DBG_DRAW) dbwin("XDrawRectangle %p x=%d y=%d width=%d height=%d pixel=%ld\n",
	(void *) d, x, y, width, height, gc->foreground);

    display->request++;

    gfxDrawRect(d, gc, x, y, width, height);

    if (IS_WINDOW(d)) {
	SdlTkScreenChanged(0);
	SdlTkDirtyAll(d);
    }
}

int
XDrawString(display, d, gc, x, y, string, length)
    Display* display;
    Drawable d;
    GC gc;
    int x;
    int y;
    CONST char* string;
    int length;
{
    gfxDrawString(d, gc, x, y, string, length);

    if (IS_WINDOW(d)) {
	SdlTkScreenChanged(0);
	SdlTkDirtyAll(d);
    }

    return 0;
}

int
XDrawString16(display, d, gc, x, y, string, length)
    Display* display;
    Drawable d;
    GC gc;
    int x;
    int y;
    CONST XChar2b* string;
    int length;
{
    gfxDrawString(d, gc, x, y, (char *) string, length);
    return 0;
}
/*
int
XEmptyRegion(r)
	Region r;
{
    return 0;
}
*/
int
XEventsQueued(display, mode)
	Display *display;
	int mode;
{
    return 0;
}

void
XFillArc(display, d, gc, x, y, width, height, start, extent)
    Display* display;
    Drawable d;
    GC gc;
    int x;
    int y;
    unsigned int width;
    unsigned int height;
    int start;
    int extent;
{
    display->request++;

    gfxFillArc(d, gc, x, y, width, height, start, extent);

    if (IS_WINDOW(d)) {
	SdlTkScreenChanged(0);
	SdlTkDirtyAll(d);
    }
}

void
XFillPolygon(display, d, gc, points, npoints, shape, mode)
    Display* display;
    Drawable d;
    GC gc;
    XPoint* points;
    int npoints;
    int shape;
    int mode;
{
    display->request++;

    gfxFillPolygon(d, gc, points, npoints, shape, mode);

    if (IS_WINDOW(d)) {
	SdlTkScreenChanged(0);
	SdlTkDirtyAll(d);
    }
}

void
XFillRectangle(display, d, gc, x, y, width, height)
    Display* display;
    Drawable d;
    GC gc;
    int x;
    int y;
    unsigned int width;
    unsigned int height;
{
    XRectangle rectangle;
    rectangle.x = x;
    rectangle.y = y;
    rectangle.width = width;
    rectangle.height = height;
    XFillRectangles(display, d, gc, &rectangle, 1);
}

void
XFillRectangles(display, d, gc, rectangles, nrectangles)
    Display* display;
    Drawable d;
    GC gc;
    XRectangle* rectangles;
    int nrectangles;
{
    int i;

    if (dbg & DBG_DRAW) dbwin("XFillRectangles n=%d d=%p pixel=%#0lx\n", nrectangles, (void *) d, gc->foreground);

    display->request++;

    for (i = 0; i < nrectangles; i++) {
	gfxFillRect(d, gc, rectangles[i].x, rectangles[i].y,
	    rectangles[i].width, rectangles[i].height);
    }

    if (IS_WINDOW(d)) {
	SdlTkScreenChanged(0);
	SdlTkDirtyAll(d);
    }
}

Bool
XFilterEvent(event, window)
    XEvent* event;
    Window window;
{
    return 0;
}

int
XFlush(display)
    Display *display;
{
    SDL_Event sdl_event;
    XEvent event;
    _XSQEvent *qevent;

    display->request++;

    /* Add all pending SDL events to the X event queue */
    while (SDL_PollEvent(&sdl_event)) {

	if (!SdlTkTranslateEvent(&sdl_event, &event))
	    continue;

	/* Grab an unused event from the list */
	qevent = display->qfree;
	if (qevent == NULL) {
	    qevent = (_XSQEvent *) ckalloc(sizeof(_XSQEvent));
	} else {
	    display->qfree = display->qfree->next;
	}

	qevent->sdl_event = sdl_event;
	qevent->event = event;
	qevent->next = NULL;

	/* Append to event queue */
	if (display->tail)
	    display->tail->next = qevent;
	else
	    display->head = qevent;
	display->tail = qevent;

	display->qlen++;

	if (dbg & DBG_EVENTS) dbwin("added 1 event to queue (%d total)\n", display->qlen);
    }

    return 0;
}

void
XForceScreenSaver(display, mode)
    Display* display;
    int mode;
{
}

int
XFree(void *data)
{
    if (data != NULL) {
	ckfree(data);
    }
    return 0;
}

void
XFreeColormap(display, colormap)
    Display* display;
    Colormap colormap;
{
}

void
XFreeColors(display, colormap, pixels, npixels, planes)
    Display* display;
    Colormap colormap;
    unsigned long* pixels;
    int npixels;
    unsigned long planes;
{
}

void
XFreeCursor(display, cursor)
    Display* display;
    Cursor cursor;
{
}

/* "The XFreeFont() function deletes the association between the font resource
ID and the specified font and frees the XFontStruct structure. The font itself
will be freed when no other resource references it. The data and the font
should not be referenced again." */
int
XFreeFont(display, font_struct)
    Display* display;
    XFontStruct* font_struct;
{
    fontFreeFont(font_struct);
    return 0;
}

int
XFreeFontNames(list)
    char** list;
{
    int i = 0;

    if (list == NULL)
	return 0;
    while (list[i] != NULL)
	ckfree(list[i++]);
    ckfree((char *) list);
    return 0;
}

void
XFreeGC(d, gc)
    Display * d;
    GC gc;
{
    d->request++;

    if (gc != None) {
	if (gc->clip_mask != None) {
	    ckfree((char*) gc->clip_mask);
	}
	memset(gc, 0xFE, sizeof(GC));
	ckfree((char *) gc);
    }

/*dbwin("FREE GC: %p\n", gc);*/
}

void
XFreeModifiermap(modmap)
    XModifierKeymap* modmap;
{
    ckfree((char *) modmap->modifiermap);
    ckfree((char *) modmap);
}

int
XFreePixmap(display, pixmap)
    Display* display;
    Pixmap pixmap;
{
    _Pixmap *_p = (_Pixmap *) pixmap;

    if (dbg & DBG_PIXMAPS) dbwin("XFreePixmap %p", (void *) _p);

    display->request++;

    if (_p == NULL)
	return 0;
    SDL_FreeSurface(_p->sdl);
    memset(_p, 0xFE, sizeof(_Pixmap));
    ckfree((char *) _p);
    return 0;
}

GContext
XGContextFromGC(gc)
    GC gc;
{
    return (GContext) NULL;
}

char *
XGetAtomName(display, atom)
    Display* display;
    Atom atom;
{
    return (char *) atom; /* see XInternAtom && XGetFontProperty */
}

Bool
XGetFontProperty(font_struct, atom, value_return)
    XFontStruct* font_struct;
    Atom atom;
    unsigned long* value_return;
{
    if (atom == XA_FONT) {
	_Font *_f = (_Font *) font_struct->fid;
	char *name;
	name = ckalloc(strlen(_f->xlfd) + 1);
	strcpy(name, _f->xlfd);
	*value_return = (unsigned long) name;
	return True;
    }
    return False;
}

Status
XGetGeometry(display, d, root_return, x_return, y_return, width_return,
	height_return, border_width_return, depth_return)
    Display* display;
    Drawable d;
    Window* root_return;
    int* x_return;
    int* y_return;
    unsigned int* width_return;
    unsigned int* height_return;
    unsigned int* border_width_return;
    unsigned int* depth_return;
{
    _Pixmap *_p = (_Pixmap *) d;
    _Window *_w = (_Window *) d;

    display->request++;

    *root_return = x_screen->root;
    
    if (_p->type == DT_PIXMAP) {
	*x_return = *y_return = 0;
	*width_return = _p->sdl->w;
	*height_return = _p->sdl->h;
	*border_width_return = 0;
	*depth_return = _p->sdl->format->BitsPerPixel;
    }
    if (_w->type == DT_WINDOW) {
	*x_return = _w->atts.x;
	*y_return = _w->atts.y;
	*width_return = _w->atts.width;
	*height_return = _w->atts.height;
	*border_width_return = _w->atts.border_width;
	*depth_return = x_screen->root_depth;
    }
    
    return 1;
}

XImage *
XGetImage(display, d, x, y, width, height, plane_mask, format)
    Display* display;
    Drawable d;
    int x;
    int y;
    unsigned int width;
    unsigned int height;
    unsigned long plane_mask;
    int format;
{
    SDL_Surface *sdl;
    char *pixels;
    _Pixmap _p;
    XGCValues fakeGC;

    /* Allocate a block of pixels to hold the result. */
    pixels = ckalloc(width * height * sdl_screen->format->BytesPerPixel);

    /* Create an SDL surface that uses the pixels we allocated above */
    sdl = SDL_CreateRGBSurfaceFrom(pixels, width, height,
	sdl_screen->format->BitsPerPixel,
	width * sdl_screen->format->BytesPerPixel, /* pitch */
	sdl_screen->format->Rmask,
	sdl_screen->format->Gmask,
	sdl_screen->format->Bmask,
	sdl_screen->format->Amask);

    if (sdl == NULL) {
	ckfree(pixels);
	return NULL;
    }

    /* Create a pixmap from the SDL surface. */
    _p.type = DT_PIXMAP;
    _p.sdl = sdl;
    _p.format = SdlTkPixelFormat(sdl);

    fakeGC.clip_mask = None;
    fakeGC.graphics_exposures = False;

    /* Copy from the drawable to our pixmap */
    gfxCopyArea(d, (Pixmap) &_p, &fakeGC, x, y, width, height, 0, 0);

    /* Free the surface. The pixels are *not* freed */
    SDL_FreeSurface(sdl);

    /* Allocate the XImage using the pixels we allocated above */ 
    return XCreateImage(display, x_screen->root_visual,
	sdl_screen->format->BitsPerPixel, ZPixmap, 0, pixels,
	width, height, 0, 0);
}

int
XGetInputFocus(display, focus_return, revert_to_return)
    Display *display;
    Window *focus_return;
    int *revert_to_return;
{
    display->request++;
    *focus_return = x_focus_window;
    *revert_to_return = RevertToParent;
    return 0;
}

XModifierKeymap	*
XGetModifierMapping(display)
    Display* display;
{
    XModifierKeymap *map = (XModifierKeymap *)ckalloc(sizeof(XModifierKeymap));

    map->max_keypermod = 1;
    map->modifiermap = (KeyCode *) ckalloc(sizeof(KeyCode)*8);
    map->modifiermap[ShiftMapIndex] = SDLK_LSHIFT;
    map->modifiermap[LockMapIndex] = SDLK_CAPSLOCK;
    map->modifiermap[ControlMapIndex] = SDLK_LCTRL;
    map->modifiermap[Mod1MapIndex] = SDLK_NUMLOCK;
    map->modifiermap[Mod2MapIndex] = 0 /*VK_MENU*/;
    map->modifiermap[Mod3MapIndex] = SDLK_SCROLLOCK;
    map->modifiermap[Mod4MapIndex] = 0;
    map->modifiermap[Mod5MapIndex] = 0;
    return map;
}

int
XGetWindowAttributes(display, w, window_attributes_return)
    Display* display;
    Window w;
    XWindowAttributes* window_attributes_return;
{
    return 0;
}

int
XGetWindowProperty(display, w, property, long_offset, long_length, delete,
	req_type, actual_type_return, actual_format_return, nitems_return,
	bytes_after_return, prop_return)
    Display* display;
    Window w;
    Atom property;
    long long_offset;
    long long_length;
    Bool delete;
    Atom req_type;
    Atom* actual_type_return;
    int* actual_format_return;
    unsigned long* nitems_return;
    unsigned long* bytes_after_return;
    unsigned char** prop_return;
{
    *actual_type_return = None;
    *actual_format_return = 0;
    *nitems_return = 0;
    *bytes_after_return = 0;
    *prop_return = NULL;
    return BadValue;
}

XVisualInfo *XGetVisualInfo(display, vinfo_mask, vinfo_template, nitems_return)
    Display* display;
    long vinfo_mask;
    XVisualInfo* vinfo_template;
    int* nitems_return;
{
    XVisualInfo *info = (XVisualInfo *)ckalloc(sizeof(XVisualInfo));

    info->visual = DefaultVisual(display, 0);
    info->visualid = info->visual->visualid;
    info->screen = 0;
    info->depth = info->visual->bits_per_rgb;
    info->class = info->visual->class;
    info->colormap_size = info->visual->map_entries;
    info->bits_per_rgb = info->visual->bits_per_rgb;
    info->red_mask = info->visual->red_mask;
    info->green_mask = info->visual->green_mask;
    info->blue_mask = info->visual->blue_mask;
    
    if (((vinfo_mask & VisualIDMask)
	    && (vinfo_template->visualid != info->visualid))
	    || ((vinfo_mask & VisualScreenMask)
		    && (vinfo_template->screen != info->screen))
	    || ((vinfo_mask & VisualDepthMask)
		    && (vinfo_template->depth != info->depth))
	    || ((vinfo_mask & VisualClassMask)
		    && (vinfo_template->class != info->class))
	    || ((vinfo_mask & VisualColormapSizeMask)
		    && (vinfo_template->colormap_size != info->colormap_size))
	    || ((vinfo_mask & VisualBitsPerRGBMask)
		    && (vinfo_template->bits_per_rgb != info->bits_per_rgb))
	    || ((vinfo_mask & VisualRedMaskMask)
		    && (vinfo_template->red_mask != info->red_mask))
	    || ((vinfo_mask & VisualGreenMaskMask)
		    && (vinfo_template->green_mask != info->green_mask))
	    || ((vinfo_mask & VisualBlueMaskMask)
		    && (vinfo_template->blue_mask != info->blue_mask))
	) {
	ckfree((char *) info);
	return NULL;
    }

    *nitems_return = 1;
    return info;
}

Status
XGetWMColormapWindows(display, w, windows_return, count_return)
    Display* display;
    Window w;
    Window** windows_return;
    int* count_return;
{
    return (Status) NULL;
}

int
XGrabKeyboard(display, grab_window, owner_events, pointer_mode,
	keyboard_mode, time)
    Display* display;
    Window grab_window;
    Bool owner_events;
    int pointer_mode;
    int keyboard_mode;
    Time time;
{
    return GrabSuccess;
}

/* tkPointer.c */
#if 0
int
XGrabPointer(d, w1, b, ui, i1, i2, w2, c, t)
    Display* d;
    Window w1;
    Bool b;
    unsigned int ui;
    int i1;
    int i2;
    Window w2;
    Cursor c;
    Time t;
{
    return 0;
}
#endif

int
XGrabServer(display)
    Display *display;
{
    return 0;
}

int
XIconifyWindow(display, w, screen_number)
    Display* display;
    Window w;
    int screen_number;
{
    return 0;
}

Atom XInternAtom(display, atom_name, only_if_exists)
    Display* display;
    _Xconst char* atom_name;
    Bool only_if_exists;
{
    char *buf = ckalloc(strlen(atom_name) + 1); /* memory leak? */
    strcpy(buf, atom_name);
    return (Atom) buf;
}
/*
void
XIntersectRegion(sra, srb, dr_return)
    Region sra;
    Region srb;
    Region dr_return;
{
}
*/
KeySym
XKeycodeToKeysym(display, keycode, index)
    Display* display;
    unsigned int keycode;
    int index;
{
    /* keycode is SDLK_xxx */
    KeySym keymap[SDLK_LAST];
    int i;

    for (i = 0; i < SDLK_LAST; i++)
	keymap[i] = NoSymbol;

    for (i = 0; i <= 26; i++) {
	keymap[SDLK_a + i] = XK_a + i;
	keymap[65 + i] = XK_A + i;
    }

    keymap[123] = XK_braceleft;
    keymap[124] = XK_bar;
    keymap[125] = XK_braceright;
    keymap[126] = XK_asciitilde;

    keymap[SDLK_SPACE] = XK_space;
    keymap[SDLK_EXCLAIM] = XK_exclam;
    keymap[SDLK_QUOTEDBL] = XK_quotedbl;
    keymap[SDLK_HASH] = XK_numbersign;
    keymap[SDLK_DOLLAR] = XK_dollar;
    keymap[37] = XK_percent;
    keymap[SDLK_AMPERSAND] = XK_ampersand;
    keymap[SDLK_QUOTE] = XK_apostrophe;
    keymap[SDLK_LEFTPAREN] = XK_parenleft;
    keymap[SDLK_RIGHTPAREN] = XK_parenright;
    keymap[SDLK_ASTERISK] = XK_asterisk;
    keymap[SDLK_PLUS] = XK_plus;
    keymap[SDLK_COMMA] = XK_comma;
    keymap[SDLK_MINUS] = XK_minus;
    keymap[SDLK_PERIOD] = XK_period;
    keymap[SDLK_SLASH] = XK_slash;

    for (i = 0; i <= 9; i++)
	keymap[SDLK_0 + i] = XK_0 + i;

    keymap[SDLK_COLON] = XK_colon;
    keymap[SDLK_SEMICOLON] = XK_semicolon;
    keymap[SDLK_LESS] = XK_less;
    keymap[SDLK_EQUALS] = XK_equal;
    keymap[SDLK_GREATER] = XK_greater;
    keymap[SDLK_QUESTION] = XK_question;
    keymap[SDLK_AT] = XK_at;

    keymap[SDLK_LEFTBRACKET] = XK_bracketleft;
    keymap[SDLK_BACKSLASH] = XK_backslash;
    keymap[SDLK_RIGHTBRACKET] = XK_bracketright;
    keymap[SDLK_CARET] = XK_asciicircum;
    keymap[SDLK_UNDERSCORE] = XK_underscore;
    keymap[SDLK_BACKQUOTE] = XK_grave;

    for (i = 0; i <= 9; i++)
	keymap[SDLK_KP0 + i] = XK_KP_0 + i;
    keymap[SDLK_KP_PERIOD] = XK_KP_Decimal;
    keymap[SDLK_KP_DIVIDE] = XK_KP_Divide;
    keymap[SDLK_KP_MULTIPLY] = XK_KP_Multiply;
    keymap[SDLK_KP_MINUS] = XK_KP_Subtract;
    keymap[SDLK_KP_PLUS] = XK_KP_Add;
    keymap[SDLK_KP_ENTER] = XK_KP_Enter;
    keymap[SDLK_KP_EQUALS] = XK_KP_Equal;

    keymap[SDLK_LSUPER] = XK_Win_L;
    keymap[SDLK_RSUPER] = XK_Win_R;
    keymap[SDLK_MENU] = XK_App;

    keymap[SDLK_BACKSPACE] = XK_BackSpace;
    keymap[SDLK_DELETE] = XK_Delete;
    keymap[SDLK_TAB] = XK_Tab;
    keymap[SDLK_RETURN] = XK_Return;
    keymap[SDLK_LALT] = XK_Alt_L;
    keymap[SDLK_LCTRL] = XK_Control_L;
    keymap[SDLK_LSHIFT] = XK_Shift_L;
    keymap[SDLK_RALT] = XK_Alt_R;
    keymap[SDLK_RCTRL] = XK_Control_R;
    keymap[SDLK_RSHIFT] = XK_Shift_R;
    keymap[SDLK_PAUSE] = XK_Pause;
    keymap[SDLK_ESCAPE] = XK_Escape;
    keymap[SDLK_PAGEUP] = XK_Prior;
    keymap[SDLK_PAGEDOWN] = XK_Next;
    keymap[SDLK_END] = XK_End;
    keymap[SDLK_HOME] = XK_Home;
    keymap[SDLK_LEFT] = XK_Left;
    keymap[SDLK_RIGHT] = XK_Right;
    keymap[SDLK_UP] = XK_Up;
    keymap[SDLK_DOWN] = XK_Down;
    keymap[SDLK_INSERT] = XK_Insert;
    for (i = 0; i < 15; i++)
	keymap[SDLK_F1 + i] = XK_F1 + i;
    keymap[SDLK_CAPSLOCK] = XK_Caps_Lock;
    keymap[SDLK_NUMLOCK] = XK_Num_Lock;
    keymap[SDLK_SCROLLOCK] = XK_Scroll_Lock;
    return keymap[keycode];
}

KeyCode
XKeysymToKeycode(display, keysym)
    Display* display;
    KeySym keysym;
{
    return 0;
}

char *
XKeysymToString(keysym)
    KeySym keysym;
{
    return NULL;
}

char **
XListFonts(display, pattern, maxnames, actual_count_return)
    Display* display;
    CONST char* pattern;
    int maxnames;
    int* actual_count_return;
{
    if (dbg & DBG_FONTS) dbwin("XListFonts %s\n", pattern);

    return ListFonts(pattern, actual_count_return);
}

XHostAddress *
XListHosts(display, nhosts_return, state_return)
    Display* display;
    int* nhosts_return;
    Bool* state_return;
{
    return NULL;
}

Font
XLoadFont(display, name)
    Display* display;
    _Xconst char* name; 
{
    return fontLoadXLFD(name);
}

XFontStruct *
XLoadQueryFont(display, name)
    Display* display;
    CONST char* name;
{
    Font f = fontLoadXLFD(name);
    return f ? ((_Font *) f)->font_struct : NULL;
}

int
XLookupColor(display, colormap, color_name, exact_def_return,
	screen_def_return)
    Display* display;
    Colormap colormap;
    _Xconst char* color_name;
    XColor* exact_def_return;
    XColor* screen_def_return;
{
    return 0;
}

/* Needed for TkpGetString */
int
XLookupString(event_struct, buffer_return, bytes_buffer, keysym_return,
    status_in_out)
    XKeyEvent* event_struct;
    char* buffer_return;
    int bytes_buffer;
    KeySym* keysym_return;
    XComposeStatus* status_in_out;
{
    /* Only care about buffer and bytes_buffer */
    /* Already converted to UTF-8 by SdlTkTranslateEvent */
    memcpy(buffer_return, event_struct->trans_chars, event_struct->nbytes);
    return event_struct->nbytes; /* length */
}

void
XMapWindow(display, w)
    Display* display;
    Window w;
{
    _Window *_w = (_Window *) w;
    XEvent event;

    display->request++;

    if (dbg & DBG_WINDOWS) dbwin("XMapWindow: %s\n", SdlTkDescribeWindow(w));

    if (_w->atts.map_state != IsUnmapped)
	return;

    /*
     * A reparenting window manager like twm will get a MapRequest event
     * when XMapWindow is called. It will then create a decorative frame
     * window to contain the window, and reparent the window inside the
     * decorative frame.
     */
    if (PARENT_IS_ROOT(w) &&
	    (_w->tkwin != NULL) &&
	    !_w->atts.override_redirect &&
	    (_w->dec == NULL)) {
	Window wdec;
	int width, height;

	if (dbg & DBG_WINDOWS) dbwin("XMapWindow: found toplevel, creating decorative frame\n");

	/* each reparent gets a drawing surface for itself and all its
	 * children */
	width = _w->atts.width + DEC_FRAME_WIDTH * 2;
	height = _w->atts.height + DEC_TITLE_HEIGHT + DEC_FRAME_WIDTH;

	wdec = XCreateWindow(display, x_screen->root,
	    _w->atts.x /*- DEC_FRAME_WIDTH*/, _w->atts.y /*- DEC_TITLE_HEIGHT*/,
	    width, height, 0, x_screen->root_depth, InputOutput,
	    x_screen->root_visual, 0, NULL);

	SdlTkResizeWindowSurface((_Window *) wdec);

	decCreate((_Window *) wdec);

	/* Reparent the window into the decorative frame, and move it
	 * to the proper place */
	XReparentWindow(display, w, wdec, DEC_FRAME_WIDTH, DEC_TITLE_HEIGHT);

	/* Let Tk know I moved the window. You would think the <Reparent>
	 * event would be a good clue */
	if (_w->atts.your_event_mask & StructureNotifyMask)
	    SdlTkGenerateConfigureNotify(display, w);
    }

    /* Map decorative frame */
    if (_w->parent->dec != NULL) {
	_w->parent->atts.map_state = IsViewable;
    }

    _w->atts.map_state = IsViewable;

    /* Tk only cares about this for wrapper windows */
    if (_w->atts.your_event_mask & StructureNotifyMask) {
	event.type = MapNotify;
	event.xmap.serial = display->request;
	event.xmap.send_event = False;
	event.xmap.display = display;
	event.xmap.event = w;
	event.xmap.window = w;
	event.xmap.override_redirect = _w->atts.override_redirect;
	Tk_QueueWindowEvent(&event, TCL_QUEUE_TAIL);
    }

    if (_w->parent->dec != NULL)
	SdlTkVisRgnChanged(_w->parent, VRC_CHANGED | VRC_DO_PARENT, 0, 0);
    else
	SdlTkVisRgnChanged(_w, VRC_CHANGED | VRC_DO_PARENT, 0, 0);

    /* Generate a <FocusIn> if this is the top-most Tk wrapper window */
    /* Don't focus on override_redirect's though (i.e., menus) */
    if (_w == SdlTkTopVisibleWrapper() && _w->parent->dec != NULL)
	XSetInputFocus(display, w, RevertToParent, CurrentTime);

    SdlTkScreenChanged(1);
}

void
XMoveWindow(display, w, x, y)
    Display* display;
    Window w;
    int x;
    int y;
{
    _Window *_w = (_Window *) w;
    int ox, oy;

    if (dbg & DBG_WINDOWS) dbwin("XMoveWindow: %s x=%d y=%d\n", SdlTkDescribeWindow(w), x, y);

    display->request++;

    /* If the window has a decorative frame, move the decorative frame
     * instead of the given window */
    if (_w->parent->dec != NULL) {
	_Window *wdec = _w->parent;
	ox = wdec->atts.x, oy = wdec->atts.y;
	wdec->atts.x = x;
	wdec->atts.y = y;
    } else {
	ox = _w->atts.x, oy = _w->atts.y;
	_w->atts.x = x;
	_w->atts.y = y;
    }

    if (_w->atts.your_event_mask & StructureNotifyMask)
	SdlTkGenerateConfigureNotify(display, w);

    if (_w->parent->dec != NULL)
	SdlTkVisRgnChanged(_w->parent, VRC_CHANGED | VRC_DO_PARENT | VRC_MOVE, ox, oy);
    else
	SdlTkVisRgnChanged(_w, VRC_CHANGED | VRC_DO_PARENT | VRC_MOVE, ox, oy);

    /* Dragging a toplevel doesn't affect visible regions of double-buffered
     * windows or their descendants */
    SdlTkScreenChanged(1 /*_w->parent->dec == NULL*/);
}

void
XMoveResizeWindow(display, w, x, y, width, height)
    Display* display;
    Window w;
    int x;			/* Position relative to parent. */
    int y;
    unsigned int width;
    unsigned int height;
{
    _Window *_w = (_Window *) w;
    int ox = 0, oy = 0, flags = VRC_CHANGED | VRC_DO_PARENT;

    display->request++;

    if (width < 1)
	width = 1;
    if (height < 1)
	height = 1;

    if (dbg & DBG_WINDOWS) dbwin("XMoveResizeWindow: %s x=%d y=%d width=%d height=%d\n", SdlTkDescribeWindow(w), x, y, width, height);

    /* If this window has a decorative frame, move the decorative frame,
     * not the given window */
    if (_w->parent->dec != NULL) {
	_Window *wdec = _w->parent;
	if (x != wdec->atts.x || y != wdec->atts.y) {
	    ox = wdec->atts.x, oy = wdec->atts.y;
	    flags |= VRC_MOVE;
	    wdec->atts.x = x;
	    wdec->atts.y = y;
	}
    } else {

	/* ConfigureEvent will call this on the children of a wrapper even if
	* their size/position doesn't change. ConfigureEvent doesn't wait
	* for <ConfigureNotify> so do nothing in this case. */
	if (x == _w->atts.x && y == _w->atts.y &&
	    width == _w->atts.width && height == _w->atts.height)
	    return;

	if (x != _w->atts.x || y != _w->atts.y) {
	    ox = _w->atts.x, oy = _w->atts.y;
	    flags |= VRC_MOVE;
	    _w->atts.x = x;
	    _w->atts.y = y;
	}
    }

    /* "wm geom +x+y will call this, even though the size doesn't change */
    if (_w->atts.width != width || _w->atts.height != height) {

	/* If this window has a decorative frame, resize it */
	if (_w->parent->dec != NULL) {
	    _Window *wdec = _w->parent;

	    wdec->atts.width = width + DEC_FRAME_WIDTH * 2;
	    wdec->atts.height = height + DEC_TITLE_HEIGHT + DEC_FRAME_WIDTH;

	    /* A window's requested width/height are *inside* its borders */
	    /* A child window is clipped within its parent's borders */
	    wdec->parentWidth = wdec->atts.width + 2 * wdec->atts.border_width;
	    wdec->parentHeight = wdec->atts.height + 2 * wdec->atts.border_width;

	    /* Resize decorative frame drawing surface */
	    SdlTkResizeWindowSurface(wdec);
	}

	_w->atts.width = width;
	_w->atts.height = height;

	/* A window's requested width/height are *inside* its borders */
	/* A child window is clipped within its parent's borders */
	_w->parentWidth = width + 2 * _w->atts.border_width;
	_w->parentHeight = height + 2 * _w->atts.border_width;

	/* Resize override_redirect drawing surface */
	if (_w->sdl != NULL)
	    SdlTkResizeWindowSurface(_w);
    }

    if (_w->atts.your_event_mask & StructureNotifyMask)
	SdlTkGenerateConfigureNotify(display, w);

    if (_w->parent->dec != NULL)
	SdlTkVisRgnChanged(_w->parent, flags, ox, oy);
    else
	SdlTkVisRgnChanged(_w, flags, ox, oy);

    SdlTkScreenChanged(1);
}

int
XNextEvent(display, event_return)
    Display* display;
    XEvent* event_return;
{
    _XSQEvent *qevent = display->head;

    if (qevent != NULL) {
	*event_return = qevent->event;

	/* Remove from front of queue */
	display->head = qevent->next;
	if (display->head == NULL)
	    display->tail = NULL;

	/* Add to front of free list */
	qevent->next = display->qfree;
	display->qfree = qevent;

	display->qlen--;

	if (dbg & DBG_EVENTS) dbwin("removed 1 event from queue (%d total)\n", display->qlen);
    }
    return 0;
}

int
XNoOp(display)
    Display *display;
{
    display->request++;
    return 0;
}

Display *
XOpenDisplay(display_name)
    CONST char* display_name;
{
    Display *display;
    Screen *screen;
    int width, height;
    Uint8  video_bpp;
    Uint32 videoflags;
    _Window *_w;

    if (x_display != NULL)
	return x_display;

    /* From cloids.cpp */
    if (SDL_Init(SDL_INIT_VIDEO) < 0) {
	fprintf(stderr, "Couldn't initialize SDL: %s", SDL_GetError());
	return NULL;
    }
    atexit(SDL_Quit);

    /* If you set SDL_HWSURFACE, it succeeds only when SDL_FULLSCREEN
    * is set. If you set SDL_HWSURFACE, you will get flickering unless
    * you use SDL_DOUBLEBUF. If you do use SDL_DOUBLEBUF then you
    * must call SDL_Flip() instead of SDL_UpdateRect(). */

    videoflags = SDL_SWSURFACE /* SDL_FULLSCREEN | SDL_HWSURFACE */ /*| SDL_ANYFORMAT*/;
    width = 1024;
    height = 768;

    {
	const SDL_VideoInfo *info = SDL_GetVideoInfo();
	SDL_PixelFormat *format = info->vfmt;
printf("BEST FORMAT:\n  BitsPerPixel %d\n  BytesPerPixel %d\n  Rmask %#0X Gmask %#0X Bmask %#0X\n",
    format->BitsPerPixel, format->BytesPerPixel,
    format->Rmask, format->Gmask, format->Bmask);
	video_bpp = format->BitsPerPixel;
    }
/*    video_bpp = 24;*/

    sdl_screen = SDL_SetVideoMode(width, height, video_bpp, videoflags);
    if (sdl_screen == NULL) {
	fprintf(stderr, "couldn't set %dx%d video mode: %s\n",
		width, height, SDL_GetError());
	fflush(stderr);
	return NULL;
    }

printf("SCREEN INFO:\n  width %d\n  height %d\n  pitch %d\n  BitsPerPixel %d\n  BytesPerPixel %d\n  Rmask %#0X Gmask %#0X Bmask %#0X\n  Rshift %u Gshift %u Bshift %u\n",
    sdl_screen->w , sdl_screen->h , sdl_screen->pitch,
    sdl_screen->format->BitsPerPixel, sdl_screen->format->BytesPerPixel,
    sdl_screen->format->Rmask, sdl_screen->format->Gmask, sdl_screen->format->Bmask,
    sdl_screen->format->Rshift, sdl_screen->format->Gshift, sdl_screen->format->Bshift);

    /* Tell SDL to give us the unicode rep of a keypress */
    SDL_EnableUNICODE(1);

    /* Tell SDL to generate repeated KeyDown events */
    SDL_EnableKeyRepeat(SDL_DEFAULT_REPEAT_DELAY, SDL_DEFAULT_REPEAT_INTERVAL);

    /* Set title for window */
    SDL_WM_SetCaption(display_name, display_name);

    /* From win/tkWinX.c TkpOpenDisplay */
    display = (Display *) ckalloc(sizeof(Display));
    memset(display, '\0', sizeof(Display));

    display->display_name = (char *) ckalloc(strlen(display_name)+1);
    strcpy(display->display_name, display_name);

    display->cursor_font = 1;
    display->nscreens    = 1;
    display->request     = 1;
    display->qlen        = 0;

    /* See TkUnixDoOneEvent */
    display->fd = 1;

    screen = (Screen *) ckalloc(sizeof(Screen));
    memset(screen, '\0', sizeof(Screen));
    screen->display = display;

    screen->white_pixel = SDL_MapRGB(sdl_screen->format, 255, 255, 255);
    screen->black_pixel = SDL_MapRGB(sdl_screen->format, 0, 0, 0);
    screen->cmap = None;

    display->screens = screen;
    display->nscreens = 1;
    display->default_screen = 0;

    /* TkWinDisplayChanged */
    screen->width = width;
    screen->height = height;
    screen->mwidth = (screen->width * 254 + 360) / 720; /* from Mac Tk */
    screen->mheight = (screen->height * 254 + 360) / 720;

    screen->root_depth = sdl_screen->format->BitsPerPixel;

    screen->root_visual = (Visual *) ckalloc(sizeof(Visual));
    screen->root_visual->visualid = 0;

    if (sdl_screen->format->palette != NULL) {
	screen->root_visual->map_entries = sdl_screen->format->palette->ncolors;
	screen->root_visual->class = PseudoColor;
	screen->root_visual->red_mask = 0x0;
	screen->root_visual->green_mask = 0x0;
	screen->root_visual->blue_mask = 0x0;
    } else if (screen->root_depth == 4) {
	screen->root_visual->class = StaticColor;
	screen->root_visual->map_entries = 16;
    } else if (screen->root_depth == 8) {
	screen->root_visual->class = StaticColor;
	screen->root_visual->map_entries = 256;
    } else if (screen->root_depth == 12) {
	screen->root_visual->class = TrueColor;
	screen->root_visual->map_entries = 32;
	screen->root_visual->red_mask = 0xf0;
	screen->root_visual->green_mask = 0xf000;
	screen->root_visual->blue_mask = 0xf00000;
    } else if (screen->root_depth == 16) {
	screen->root_visual->class = TrueColor;
	screen->root_visual->map_entries = 64;
	screen->root_visual->red_mask = sdl_screen->format->Rmask;
	screen->root_visual->green_mask = sdl_screen->format->Gmask;
	screen->root_visual->blue_mask = sdl_screen->format->Bmask;
    } else if (screen->root_depth >= 24) {
	screen->root_visual->class = TrueColor;
	screen->root_visual->map_entries = 256;
	screen->root_visual->red_mask = sdl_screen->format->Rmask;
	screen->root_visual->green_mask = sdl_screen->format->Gmask;
	screen->root_visual->blue_mask = sdl_screen->format->Bmask;
    }
    screen->root_visual->bits_per_rgb = sdl_screen->format->BitsPerPixel;

    screen->cmap = XCreateColormap(display, None, screen->root_visual,
	    AllocNone);

    /* Create the root (desktop) window */
    _w = (_Window *) ckalloc(sizeof(_Window));
    memset(_w, 0, sizeof(_Window));
    _w->type = DT_WINDOW;
    _w->sdl = sdl_screen;
    _w->format = SdlTkPixelFormat(sdl_screen);
    _w->atts.x = 0;
    _w->atts.y = 0;
    _w->atts.width = _w->parentWidth = sdl_screen->w;
    _w->atts.height = _w->parentHeight = sdl_screen->h;
    _w->atts.border_width = 0;
    _w->atts.map_state = IsViewable;
    _w->visRgn = RgnPoolGet();
    _w->visRgnInParent = RgnPoolGet();
    if (doublebuffer_windows)
	_w->visRgnInParent2 = RgnPoolGet();
    screen->root = (Window) _w;

    if (dbg & DBG_WINDOWS) dbwin("created root window %p\n", (void *) _w);

    /* Nasty globals */
    _root = _w;
    x_display = display;
    x_screen = screen;

    if (fontInit() != TCL_OK)
	return NULL;

    return display;
}

void
XPutBackEvent(display, event)
    Display* display;
    XEvent* event;
{
}

int
XPutImage(display, d, gc, image, src_x, src_y, dest_x, dest_y, width, height)
    Display* display;
    Drawable d;
    GC gc;
    XImage* image;
    int src_x;
    int src_y;
    int dest_x;
    int dest_y;
    unsigned int width;
    unsigned int height;
{
    gfxPutImage(d, image, src_x, src_y, dest_x, dest_y, width, height);
    return 0;
}

/* TkTreeCtrl loupe uses this */
void
XQueryColors(display, colormap, defs_in_out, ncolors)
    Display* display;
    Colormap colormap;
    XColor* defs_in_out;
    int ncolors;
{
    int i;
    Uint32 rm, gm, bm;
    int rs, gs, bs;

    rm = sdl_screen->format->Rmask;
    gm = sdl_screen->format->Gmask;
    bm = sdl_screen->format->Bmask;

    rs = sdl_screen->format->Rshift;
    gs = sdl_screen->format->Gshift;
    bs = sdl_screen->format->Bshift;

    for (i = 0; i < ncolors; i++) {
	defs_in_out[i].red = ((defs_in_out[i].pixel & rm) >> rs) / 255.0 * USHRT_MAX;
	defs_in_out[i].green = ((defs_in_out[i].pixel & gm) >> gs) / 255.0 * USHRT_MAX;
	defs_in_out[i].blue = ((defs_in_out[i].pixel & bm) >> bs) / 255.0 * USHRT_MAX;
    }
}

Bool
XQueryPointer(display, w, root_return, child_return, root_x_return,
	root_y_return, win_x_return, win_y_return, mask_return)
    Display* display;
    Window w;
    Window* root_return;
    Window* child_return;
    int* root_x_return;
    int* root_y_return;
    int* win_x_return;
    int* win_y_return;
    unsigned int* mask_return;
{
    int state;

    display->request++;

    *mask_return = 0;

    state = SDL_GetMouseState(root_x_return, root_y_return);
    *win_x_return = *root_x_return;
    *win_y_return = *root_y_return;

    if (state & SDL_BUTTON(1))
	*mask_return |= Button1Mask;
    if (state & SDL_BUTTON(2))
	*mask_return |= Button2Mask;
    if (state & SDL_BUTTON(3))
	*mask_return |= Button3Mask;
#if 0
    if (sdl_event->key.keysym.mod & KMOD_CAPS)
	*mask_return |= LockMask;
    if (sdl_event->key.keysym.mod & KMOD_CTRL)
	*mask_return |= ControlMask;
    if (sdl_event->key.keysym.mod & KMOD_NUM)
	*mask_return |= Mod1Mask;
    if (sdl_event->key.keysym.mod & KMOD_SHIFT)
	*mask_return |= ShiftMask;
#endif
    return True;
}

int
XQueryTree(display, w, root_return, parent_return, children_return,
	nchildren_return)
    Display* display;
    Window w;
    Window* root_return;
    Window* parent_return;
    Window** children_return;
    unsigned int* nchildren_return;
{
    _Window *_w = (_Window *) w;
    _Window *child;
    int n = 0;

    display->request++;

    *root_return = x_screen->root;
    *parent_return = (Window) _w->parent;

    if (_w->child == NULL) {
	*nchildren_return = 0;
	return 1;
    }

    /* Count children */
    child = _w->child;
    while (child != NULL) {
	n++;
	child = child->next;
    }

    /* Make array of children */
    *children_return = (Window *) ckalloc(sizeof(Window) * n);
    n = 0;
    child = _w->child;
    while (child != NULL) {
	(*children_return)[n++] = (Window) child;
	child = child->next;
    }
    *nchildren_return = n;

    return 1;
}

/*
The XReconfigureWMWindow() function issues a ConfigureWindow request on the
specified top-level window. If the stacking mode is changed and the request
fails with a BadMatch error, the error is trapped by Xlib and a synthetic
ConfigureRequestEvent containing the same configuration parameters is sent
to the root of the specified window. Window managers may elect to receive
this event and treat it as a request to reconfigure the indicated window.
It returns a nonzero status if the request or event is successfully sent;
otherwise, it returns a zero status.
*/
/* Called by TkWmRestackToplevel */
Status
XReconfigureWMWindow(display, w, screen_number, mask, changes)
    Display* display;
    Window w;
    int screen_number;
    unsigned int mask;
    XWindowChanges* changes;
{
    _Window *_w = (_Window *) w;
    _Window *parent = _w->parent;
    _Window *sibling = NULL;

    display->request++;

    if (mask & CWStackMode) {

	SdlTkScreenChanged(1);

	/* Attempting to restack a wrapper? Restack decframe instead. */
	/* override_redirects won't have a decframe however */
	if (parent->dec != NULL) {
	    _w = parent;
	    parent = parent->parent;
	}

	/* Stack above/below decframe of sibling if any */
	if (mask & CWSibling) {
	    sibling = (_Window *) changes->sibling;
	    if (sibling->parent->dec != NULL) {
		sibling = sibling->parent;
	    }
	}

	SdlTkRestackWindow(_w, sibling, changes->stack_mode);
	SdlTkRestackTransients(_w);
    }

    return 0;
}
/*
int
XRectInRegion(r, x, y, width, height)
    Region r;
    int x;
    int y;
    unsigned int width;
    unsigned int height;
{
    return 0;
}
*/
void
XRefreshKeyboardMapping(event_map)
    XMappingEvent* event_map;
{
}

int
XReparentWindow(display, w, parent, x, y)
    Display* display;
    Window w;
    Window parent;
    int x;
    int y;
{
    _Window *_parent = (_Window *) parent;
    _Window *_w = (_Window *) w;
    XEvent event;

    if (dbg & DBG_WINDOWS) dbwin("XReparentWindow w=%s parent=%s x=%d y=%d\n", SdlTkDescribeWindow(w), SdlTkDescribeWindow(parent), x, y);

    display->request++;

    /* Remove from old parent */
    if (_w->parent != NULL)
	SdlTkRemoveFromParent(_w);

    /* Add to new parent */
    _w->parent = _parent;
    _w->next = _parent->child;
    _parent->child = _w;

    /* Update position */
    _w->atts.x = x;
    _w->atts.y = y;

    event.type = ReparentNotify;
    event.xreparent.serial = display->request;
    event.xreparent.send_event = False;
    event.xreparent.display = display;
    event.xreparent.event = w;
    event.xreparent.window = w;
    event.xreparent.parent = parent;
    event.xreparent.x = x;
    event.xreparent.y = y;
    event.xreparent.override_redirect = _w->atts.override_redirect;
    Tk_QueueWindowEvent(&event, TCL_QUEUE_TAIL);

    return 0;
}

void
XResizeWindow(display, w, width, height)
    Display* display;
    Window w;
    unsigned int width;
    unsigned int height;
{
    _Window *_w = (_Window *) w;

    display->request++;

    if (width < 1)
	width = 1;
    if (height < 1)
	height = 1;

    /* If this window has a decorative frame, resize it */
    if (_w->parent->dec != NULL) {
	_Window *wdec = _w->parent;

	wdec->atts.width = width + DEC_FRAME_WIDTH * 2;
	wdec->atts.height = height + DEC_TITLE_HEIGHT + DEC_FRAME_WIDTH;

	/* A window's requested width/height are *inside* its borders */
	/* A child window is clipped within its parent's borders */
	wdec->parentWidth = wdec->atts.width + 2 * wdec->atts.border_width;
	wdec->parentHeight = wdec->atts.height + 2 * wdec->atts.border_width;

	/* Reallocate the decorative frame drawing surface */
	SdlTkResizeWindowSurface(wdec);
    }

    _w->atts.width = width;
    _w->atts.height = height;

    /* A window's requested width/height are *inside* its borders */
    /* A child window is clipped within its parent's borders */
    _w->parentWidth = width + 2 * _w->atts.border_width;
    _w->parentHeight = height + 2 * _w->atts.border_width;

    /* Resize override_redirect drawing surface */
    if (_w->sdl != NULL)
	SdlTkResizeWindowSurface(_w);

    if (_w->atts.your_event_mask & StructureNotifyMask)
	SdlTkGenerateConfigureNotify(display, w);

    if (_w->parent->dec != NULL)
	SdlTkVisRgnChanged(_w->parent, VRC_CHANGED | VRC_DO_PARENT, 0, 0);
    else
	SdlTkVisRgnChanged(_w, VRC_CHANGED | VRC_DO_PARENT, 0, 0);

    SdlTkScreenChanged(1);
}

Window
XRootWindow(display, screen_number)
    Display* display;
    int screen_number;
{
    display->request++;
    return x_screen->root;
}

void
XSelectInput(display, w, event_mask)
    Display* display;
    Window w;
    long event_mask;
{
}

int
XSendEvent(display, w, propagate, event_mask, event_send)
    Display* display;
    Window w;
    Bool propagate;
    long event_mask;
    XEvent* event_send;
{
    return 0;
}

int
XSetCommand(display, w, argv, argc)
    Display* display;
    Window w;
    char** argv;
    int argc;
{
    return 0;
}

void 
XSetBackground(display, gc, background)
    Display *display;
    GC gc;
    unsigned long background;
{
    gc->background = background;
}

int
XSetClassHint(display, w, class_hints)
    Display* display;
    Window w;
    XClassHint* class_hints;
{
    return 0;
}

void
XSetClipMask(display, gc, pixmap)
    Display* display;
    GC gc;
    Pixmap pixmap;
{
    if (pixmap == None) {
	if (gc->clip_mask) {
	    ckfree((char*) gc->clip_mask);
	    gc->clip_mask = None;
	}
	return;
    }

    if (gc->clip_mask == None) {
	gc->clip_mask = (Pixmap)ckalloc(sizeof(TkpClipMask));
    }
    ((TkpClipMask*)gc->clip_mask)->type = TKP_CLIP_PIXMAP;
    ((TkpClipMask*)gc->clip_mask)->value.pixmap = pixmap;
}

void
XSetClipOrigin(display, gc, clip_x_origin, clip_y_origin)
    Display* display;
    GC gc;
    int clip_x_origin;
    int clip_y_origin;
{
    gc->clip_x_origin = clip_x_origin;
    gc->clip_y_origin = clip_y_origin;
}

void
XSetDashes(display, gc, dash_offset, dash_list, n)
    Display* display;
    GC gc;
    int dash_offset;
    _Xconst char* dash_list;
    int n;
{
    char *p = &(gc->dashes);

    gc->dash_offset = dash_offset;
    if (n > MAX_DASH_LIST_SIZE) n = MAX_DASH_LIST_SIZE;
    while (n-- > 0) {
	*p++ = *dash_list++;
    }
    *p = 0;
}

XErrorHandler
XSetErrorHandler(handler)
    XErrorHandler handler;
{
    return NULL;
}

void
XSetFont(display, gc, font)
    Display *display;
    GC gc;
    Font font;
{
    gc->font = font;
}

void 
XSetForeground(display, gc, foreground)
    Display *display;
    GC gc;
    unsigned long foreground;
{
    gc->foreground = foreground;
}

void
XSetIconName(display, w, icon_name)
    Display* display;
    Window w;
    _Xconst char* icon_name;
{
}

void
XSetInputFocus(display, focus, revert_to, time)
    Display* display;
    Window focus; /* always a Tk wrapper */
    int revert_to;
    Time time;
{
    XEvent event;

    if (dbg & DBG_FOCUS) dbwin("XSetInputFocus %s\n", (focus == None) ? "None" : (focus == PointerRoot) ? "PointerRoot" : SdlTkDescribeWindow(focus));

    display->request++;

    if (x_focus_window == focus)
	return;
#if 1
    if (x_focus_window != None) {
if (dbg & DBG_FOCUS) dbwin("XSetInputFocus: FocusOut %s\n", SdlTkDescribeWindow(x_focus_window));
	event.type = FocusOut;
	event.xfocus.serial = ++display->request;
	event.xfocus.send_event = False;
	event.xfocus.display = display;
	event.xfocus.window = x_focus_window;
	event.xfocus.mode = NotifyNormal;
	event.xfocus.detail = NotifyNonlinear;
	Tk_QueueWindowEvent(&event, TCL_QUEUE_TAIL);
    }
#endif
    x_focus_window = focus;

    if (focus != None && focus != PointerRoot) {
	event.type = FocusIn;
	event.xfocus.serial = ++display->request;
	event.xfocus.send_event = False;
	event.xfocus.display = display;
	event.xfocus.window = focus;
	event.xfocus.mode = NotifyNormal;
	event.xfocus.detail = NotifyNonlinear;
	Tk_QueueWindowEvent(&event, TCL_QUEUE_TAIL);
    }
}

void
XSetLineAttributes(display, gc, line_width, line_style, cap_style,
	join_style)
    Display *display;
    GC gc;
    unsigned int line_width;
    int line_style;
    int cap_style;
    int join_style;
{
    gc->line_width = line_width;
    gc->line_style = line_style;
    gc->cap_style = cap_style;
    gc->join_style = join_style;
}

int
XSetRegion(display, gc, r)
    Display* display;
    GC gc;
    Region r;
{
    if (r == None) {
	if (gc->clip_mask) {
	    ckfree((char*) gc->clip_mask);
	    gc->clip_mask = None;
	}
	return 1;
    }

    if (gc->clip_mask == None) {
	gc->clip_mask = (Pixmap)ckalloc(sizeof(TkpClipMask));
    }
    ((TkpClipMask*)gc->clip_mask)->type = TKP_CLIP_REGION;
    ((TkpClipMask*)gc->clip_mask)->value.region = (TkRegion) r;
	return 1;
}

void
XSetSelectionOwner(display, selection, owner, time)
    Display* display;
    Atom selection;
    Window owner;
    Time time;
{
}

int
XSetTransientForHint(display, w, prop_window)
    Display* display;
    Window w;
    Window prop_window;
{
    ((_Window *) w)->master = (_Window *) prop_window;
dbwin("%s master is now %s\n", SdlTkDescribeWindow(w), SdlTkDescribeWindow(prop_window));
    return 0;
}

void
XSetTSOrigin(display, gc, x, y)
    Display *display;
    GC gc;
    int x, y;
{
    gc->ts_x_origin = x;
    gc->ts_y_origin = y;
}

void
XSetWindowBackground(display, w, background_pixel)
    Display* display;
    Window w;
    unsigned long background_pixel;
{
}

void
XSetWindowBackgroundPixmap(display, w, background_pixmap)
    Display* display;
    Window w;
    Pixmap background_pixmap;
{
}

void
XSetWindowBorder(display, w, border_pixel)
    Display* display;
    Window w;
    unsigned long border_pixel;
{
}

void
XSetWindowBorderPixmap(display, w, border_pixmap)
    Display* display;
    Window w;
    Pixmap border_pixmap;
{
}

void
XSetWindowBorderWidth(display, w, width)
    Display* display;
    Window w;
    unsigned int width;
{
    _Window *_w = (_Window *) w;

    display->request++;

    _w->atts.border_width = width;
    _w->parentWidth = _w->atts.width + 2 * width;
    _w->parentHeight = _w->atts.height + 2 * width;
    SdlTkScreenChanged(1);
}

void
XSetWindowColormap(display, w, colormap)
    Display* display;
    Window w;
    Colormap colormap;
{
}

void
XSetWMClientMachine(display, w, text_prop)
    Display* display;
    Window w;
    XTextProperty* text_prop;
{
}

Status XSetWMColormapWindows(display, w, colormap_windows, count)
    Display* display;
    Window w;
    Window* colormap_windows;
    int count;
{
    return 0;
}

int
XSetWMHints(display, w, wm_hints)
    Display* display;
    Window w;
    XWMHints* wm_hints;
{
    return 0;
}

void
XSetWMNormalHints(display, w, hints)
    Display* display;
    Window w;
    XSizeHints* hints;
{
    _Window *_w = (_Window *) w;

    if (hints->flags & PBaseSize) {
	_w->size.base_width = hints->base_width;
	_w->size.base_height = hints->base_height;
    }
    if (hints->flags & PMinSize) {
	_w->size.min_width = hints->min_width;
	_w->size.min_height = hints->min_height;
    }
    if (hints->flags & PMaxSize) {
	_w->size.max_width = hints->max_width;
	_w->size.max_height = hints->max_height;
    }
    if (hints->flags & PResizeInc) {
	_w->size.width_inc = hints->width_inc;
	_w->size.height_inc = hints->height_inc;
    }
    _w->size.flags = hints->flags;
}

int XStoreName(display, w, window_name)
    Display* display;
    Window w;
    _Xconst char* window_name;
{
    return 0;
}

Status
XStringListToTextProperty(list, count, text_prop_return)
    char** list;
    int count;
    XTextProperty* text_prop_return;
{
    return (Status) NULL;
}

KeySym
XStringToKeysym(string)
    _Xconst char *string;
{
    return NoSymbol;
}
/*
void
XSubtractRegion(sra, srb, dr_return)
    Region sra;
    Region srb;
    Region dr_return;
{
}
*/
int
XSync(display, discard)
    Display* display;
    Bool discard;
{
    display->request++;
    return 0;
}

int
XSynchronize(display, discard)
    Display* display;
    Bool discard;
{
    display->request++;
    return 0;
}

int
XTextWidth(font_struct, string, count)
    XFontStruct* font_struct;
    CONST char* string;
    int count;
{
    return gfxTextWidth(font_struct->fid, string, count);
}

int
XTextWidth16(font_struct, string, count)
    XFontStruct* font_struct;
    CONST XChar2b* string;
    int count;
{
    return gfxTextWidth(font_struct->fid, (char *) string, count);
}

/*
If XTranslateCoordinates() returns True, it takes the src_x and src_y
coordinates relative to the source window's origin and returns these
coordinates to dest_x_return and dest_y_return relative to the destination
window's origin. If XTranslateCoordinates() returns False, src_w and dest_w
are on different screens, and dest_x_return and dest_y_return are zero.
If the coordinates are contained in a mapped child of dest_w, that child is
returned to child_return. Otherwise, child_return is set to None.
*/
Bool
XTranslateCoordinates(display, src_w, dest_w, src_x, src_y, dest_x_return,
	dest_y_return, child_return)
    Display* display;
    Window src_w; 
    Window dest_w;
    int src_x;
    int src_y;
    int* dest_x_return;
    int* dest_y_return;
    Window* child_return;
{
    _Window *_src = (_Window *) src_w;
    _Window *_dest = (_Window *) dest_w;
    int rootx, rooty;

    SdlTkRootCoords(_src, &rootx, &rooty);
    src_x += rootx;
    src_y += rooty;

    SdlTkRootCoords(_dest, &rootx, &rooty);
    *dest_x_return = src_x - rootx;
    *dest_y_return = src_y - rooty;

    *child_return = (Window) SdlTkPointToWindow(_dest, src_x, src_y, True);
    if (*child_return == dest_w)
	*child_return = None;

    return True;
}

void
XUngrabKeyboard(display, time)
    Display* display;
    Time time;
{
}

/* tkPointer.c */
#if 0
void
XUngrabPointer(d, t)
    Display* d;
    Time t;
{
}
#endif

int
XUngrabServer(display)
    Display *display;
{
    return 0;
}
/*
void
XUnionRectWithRegion(rectangle, src_region, dest_region_return)
    XRectangle* rectangle;
    Region src_region;
    Region dest_region_return;
{
}
*/

/*
"The XUnmapWindow() function unmaps the specified window and causes the X
server to generate an UnmapNotify event. If the specified window is already
unmapped, XUnmapWindow() has no effect. Normal exposure processing on formerly
obscured windows is performed. Any child window will no longer be visible until
another map call is made on the parent. In other words, the subwindows are
still mapped but are not visible until the parent is mapped. Unmapping a window
will generate Expose events on windows that were formerly obscured by it."
*/
/* Tk generates UnmapNotify events for all the children recursively, but I
 * can't figure out where that is done. */
void
XUnmapWindow(display, w)
    Display* display;
    Window w;
{
    _Window *_w = (_Window *) w;
    XEvent event;

    if (dbg & DBG_WINDOWS) dbwin("XUnmapWindow: %s\n", SdlTkDescribeWindow(w));

    if (_w->atts.map_state == IsUnmapped)
	return;

    display->request++;

    /* Unmap decorative frame */
    if (_w->parent->dec != NULL) {
	_w->parent->atts.map_state = IsUnmapped;
    }

    _w->atts.map_state = IsUnmapped;

    /* Tk only cares about this for wrapper windows */
    if (_w->atts.your_event_mask & StructureNotifyMask) {
	event.type = UnmapNotify;
	event.xunmap.serial = display->request;
	event.xunmap.send_event = False;
	event.xunmap.display = display;
	event.xunmap.event = w;
	event.xunmap.window = w;
	event.xunmap.from_configure = False;
	Tk_QueueWindowEvent(&event, TCL_QUEUE_TAIL);
    }

    if (_w->parent->dec != NULL)
	SdlTkVisRgnChanged(_w->parent, VRC_CHANGED | VRC_DO_PARENT, 0, 0);
    else
	SdlTkVisRgnChanged(_w, VRC_CHANGED | VRC_DO_PARENT, 0, 0);

    /* "All FocusOut events caused by a window unmap are generated after any
     * UnmapNotify event" */
    if (x_focus_window == w)
	SdlTkLostFocusWindow();

    SdlTkScreenChanged(1);
}

void
XWindowEvent(display, w, event_mask, event_return)
    Display* display;
    Window w;
    long event_mask;
    XEvent* event_return;
{
}

int
XWithdrawWindow(display, w, screen_number)
    Display* display;
    Window w;
    int screen_number;
{
    XUnmapWindow(display, w);
    return 1;
}

int
XmbLookupString(ic, event, buffer_return, bytes_buffer, keysym_return,
	status_return)
    XIC ic;
    XKeyPressedEvent* event;
    char* buffer_return;
    int bytes_buffer;
    KeySym* keysym_return;
    Status* status_return;
{
    return 0;
}

VisualID
XVisualIDFromVisual(visual)
	Visual* visual;
{
    return 0;
}

void
XWarpPointer(display, src_w, dest_w, src_x, src_y, src_width,
	src_height, dest_x, dest_y)
    Display* display;
    Window src_w;
    Window dest_w;
    int src_x;
    int src_y;
    unsigned int src_width;
    unsigned int src_height;
    int dest_x;
    int dest_y;
{
}

